@@ -1,6 +1,7 @@
 {-# LANGUAGE BangPatterns   #-}
 {-# LANGUAGE CPP            #-}
 {-# LANGUAGE EmptyDataDecls #-}
+{-# LANGUAGE ExplicitForAll #-}
 
 -- | This module provides wrappers in 'IO' around the functions from
 -- "Data.HashTable.Class".
@@ -56,21 +57,17 @@ module Data.HashTable.IO
 
 
 ------------------------------------------------------------------------------
-import           Control.Monad.Primitive  (PrimState)
-#if MIN_VERSION_base(4,4,0)
-import           Control.Monad.ST         (stToIO)
-import           Control.Monad.ST.Unsafe  (unsafeIOToST)
-#else
-import Control.Monad.ST (stToIO, unsafeIOToST)
-#endif
-import           Data.Hashable            (Hashable)
-import qualified Data.HashTable.Class     as C
-import           Prelude                  hiding (lookup, mapM_)
+import           Control.Monad.Primitive       (PrimState)
+import           Control.Monad.ST              (stToIO)
+import           Data.Hashable                 (Hashable)
+import qualified Data.HashTable.Class          as C
+import           Prelude                       hiding (lookup, mapM_)
 
 ------------------------------------------------------------------------------
-import qualified Data.HashTable.ST.Basic  as B
-import qualified Data.HashTable.ST.Cuckoo as Cu
-import qualified Data.HashTable.ST.Linear as L
+import           Data.HashTable.Internal.Utils (unsafeIOToST)
+import qualified Data.HashTable.ST.Basic       as B
+import qualified Data.HashTable.ST.Cuckoo      as Cu
+import qualified Data.HashTable.ST.Linear      as L
 
 
 ------------------------------------------------------------------------------
@@ -113,7 +110,7 @@ newSized = stToIO . C.newSized
 
 
 ------------------------------------------------------------------------------
--- | See the documentation for this function in "Data.HashTable.Class#v:update".
+-- | See the documentation for this function in "Data.HashTable.Class#v:insert".
 insert   :: (C.HashTable h, Eq k, Hashable k) =>
             IOHashTable h k v -> k -> v -> IO ()
 insert h k v = stToIO $ C.insert h k v
